/* 
 * Copyright (c) 2008 Andrea Di Pasquale <spikey.it@gmail.com>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice(s), this list of conditions and the following disclaimer as
 *    the first lines of this file unmodified other than the possible
 *    addition of one or more copyright notices.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice(s), this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDER(S) ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT HOLDER(S) BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 * CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
 * DAMAGE.
 *
 * $ArpON: arpon.c,v 1.10 05/31/2008 00:42:26 Andrea Di Pasquale Exp $
 */

#include <sys/ioctl.h>
#ifdef LINUX
#include <sys/poll.h>
#endif
#include <sys/types.h>
#include <sys/time.h>
#include <sys/socket.h>
#include <arpa/inet.h>
#include <netinet/in.h>
#ifdef LINUX
#include <netinet/ether.h>
#endif
#include <net/ethernet.h>
#include <net/if.h>

#include <pcap.h>
#include <dnet.h>
#include <libnet.h>

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <unistd.h>
#include <errno.h>
#include <time.h>
#include <getopt.h>
#include <fcntl.h>

#include "include/queue.h"

#define	CUR_VERSION	("1.10")
#define	WEB_SITE	("http://arpon.sourceforge.net")

#define ROOT_USER       (0x0)

#define	ERROR(str)	fprintf(stderr, "ERROR: %s @ %s:%d.\n", str, __FILE__, __LINE__)

#ifdef LINUX
#define octet	ether_addr_octet
#endif	/* struct ether_addr */

#define MAC_ADDR_LEN    	(0x12)
#define MAC_ADDR_BROADCAST   	("ff:ff:ff:ff:ff:ff")

#define INET4_ADDR_LEN  	(0x10)

#define NETMASK_CLASS_A (16777215)
#define NETMASK_CLASS_B (65535)
#define NETMASK_CLASS_C (255)

#define	DEVICE_MANUAL	(0x1)
#define	DEVICE_AUTO	(0x2)
#define	DEVICE_LIST	(0x4)

#define	PING_HOST		(0x1)
#define PING_BROADCAST		(0x2)
#define PING_REPLY_FALSE        (-1)
#define PING_REPLY_TRUE 	(0)

#define ARP_CACHE_ADD   (0x1)
#define ARP_CACHE_DEL   (0x2)
#define ARP_CACHE_LIST  (0x4)

#define ARP_CACHE       (0x1)
#define SARPI		(0x2)
#define DARPI           (0x4)

struct arp_header {	
	struct arp_hdr		header;		/* Arp header */
	struct arp_ethip 	addresses;	/* Ethernet header */
};

struct arp_cache {	
	TAILQ_ENTRY(arp_cache) entries;
	struct ether_addr	mac;
	struct in_addr		ip;
};

struct darpi_cache {
	TAILQ_ENTRY(darpi_cache) entries;
	struct in_addr	ip;	/* Entry Inet4 address */
};

static int	 task_mode_daemon(void);

static int	 device_manager(unsigned int, char *);
static char 	*device_get_default(void);
static void	 device_set_name(char *);
static int	 device_check_datalink(char *);
static int	 device_del_promisc(void);
static int	 device_get_mac_address(void);
static int	 device_get_inet4_addresses(void);
static void	 device_info_print(void);

static int 	 arp_ping_set_timeout(char *);
static int	 arp_ping_host(char *);
static int	 arp_ping_broadcast(void);
static pcap_t 	*arp_ping_open_packet(void);
static void	 arp_ping_close_packet(pcap_t *);
static int	 arp_ping_recv_packet(unsigned int, pcap_t *, unsigned char *);
static void	 arp_ping_read_packet(unsigned char *, const struct pcap_pkthdr *, const unsigned char *);
static int	 arp_ping_send_packet(unsigned int, struct ether_addr *, struct in_addr *);

static int	 arp_sniff_packets(void);
static void	 arp_sniff_signal(int);
static pcap_t 	*arp_sniff_open_packets(void);
static void	 arp_sniff_close_packets(pcap_t * pcap);
static void	 arp_sniff_read_packets(unsigned char *, const struct pcap_pkthdr *, const unsigned char *);

static int	 arp_cache_manager(unsigned int, char *);
static int	 arp_cache_add(unsigned int, arp_t *, char *);
static int	 arp_cache_del(unsigned int, arp_t *, char *);
static int	 arp_cache_list(arp_t *);
static int	 arp_cache_list_create(const struct arp_entry *, void *);
static void	 arp_cache_list_print(void);
static void	 arp_cache_list_destroy(void);

static void	 sarpi_set_timeout(char *);
static int	 sarpi_manager(void);
static void	 sarpi_manager_signal(int);
static arp_t 	*sarpi_arp_cache_open(void);
static void	 sarpi_arp_cache_close(arp_t *);
static void	 sarpi_arp_cache_list(void);
static int	 sarpi_arp_cache_refresh(arp_t *);
static int	 sarpi_realtime(arp_t *);
static pcap_t 	*sarpi_realtime_open_packets(void);
static void	 sarpi_realtime_close_packets(pcap_t *);
static void	 sarpi_realtime_read_packets(unsigned char *, const struct pcap_pkthdr *, const unsigned char *);

static int	 darpi_manager(void);
static arp_t 	*darpi_arp_cache_open(void);
static void	 darpi_arp_cache_close(arp_t *);
static int	 darpi_arp_cache_delete_entries(arp_t *);
static int	 darpi_arp_cache_add_entry(struct in_addr *, struct ether_addr *);
static int	 darpi_arp_cache_del_entry(char *);
static int	 darpi_realtime(arp_t *);
static void	 darpi_realtime_signal(int);
static pcap_t 	*darpi_realtime_open_packets(void);
static void	 darpi_realtime_close_packets(pcap_t *);
static void	 darpi_realtime_read_packets(unsigned char *, const struct pcap_pkthdr *, const unsigned char *);
static int	 darpi_cache_list_create(struct in_addr *);
static void	 darpi_cache_list_print(void);
static int	 darpi_cache_list_search(char *);
static void	 darpi_cache_list_destroy(void);

static void	 license(void);
static void	 version(void);
static void	 help(void);

static struct arp_header *arp_packet;

static struct arp_cache *arp_cache_begin, *arp_cache_next, *arp_cache_pos;
static TAILQ_HEAD(, arp_cache) arp_cache_head;

static struct darpi_cache *darpi_cache_begin, *darpi_cache_next, *darpi_cache_pos;
static TAILQ_HEAD(, darpi_cache) darpi_cache_head;

static pcap_t *pcap_sniff;

static arp_t *sarpi_arp, *darpi_arp;

static pid_t sarpi_pid_child;

static char device[IF_NAMESIZE];
static int device_offset;	/* offset for ethernet/wireless datalink */
static struct ether_addr device_mac;
static struct in_addr device_inet4, device_netmask;

static int	ping_timeout = 500,		/* Arp ping timeout, default 500 ms */
		ping_type,			/* Ping to host or broadcast */
		ping_check_reply = -1,		/* Check Arp reply if it's true or false */
		ping_broadcast_count = 1,	/* Active hosts count */
	   	arp_sniff_request_count = 0, 	/* Arp request count in passive sniffer mode */
	   	arp_sniff_reply_count = 0, 	/* Arp reply count in passive sniffer mode */
	   	sarpi_timeout = 10;		/* Default Arp cache refresh timeout */

/*
 * TASK MODE handler:
 */

/* Demonize the process. */
static int
task_mode_daemon()
{
	int fd;

	switch (fork()) 
	{
		case (-1):
			ERROR(strerror(errno));
			return (-1);
		
		case (0):
			break;
        	
		default:
			return (-1);
	}

	if (setsid() < 0)
	{
		ERROR(strerror(errno));
		return (-1);
	}
	
	printf("Task is going background mode with PID %d...\n", getpid());	
	if ((fd = open("/dev/null", O_RDWR, 0)) < 0)
	{
		ERROR(strerror(errno));
                return (-1);
        }

	dup2(fd, STDIN_FILENO);
	dup2(fd, STDOUT_FILENO);
		
	if (fd > 2)
		close(fd);
	
	return (0);
}

/*
 * DEVICE Handler:
 */

/* 
 * Handles the network interface with following modes:
 *     - Manual, 
 *     - Automatic, 
 *     - Listing eventual network interfaces
 *
 * doing the following operations:
 *     - Verifying Datalink 
 *     - Selecting the network interface to be used
 *     - Putting down the promiscue flag if found set
 *     - Reading MAC Address
 *     - Reading IP, netmask inet4 addresses
 *     - Printing out network interfaces dates
 */
static int
device_manager(unsigned int op_dev, char *dev)
{
	pcap_if_t *ifc;
	char errbuf[PCAP_ERRBUF_SIZE];

	if (pcap_findalldevs(&ifc, errbuf) < 0) 
	{
		ERROR(errbuf);
		return (-1);
	}	

	for (; ifc != NULL; ifc = ifc->next) 
	{
		switch (op_dev)
		{
			case (DEVICE_MANUAL):
				if (strncmp(dev, ifc->name, IF_NAMESIZE) == 0)
				{
					if ((ifc->flags & PCAP_IF_LOOPBACK) == 0) 
					{
						if (device_check_datalink(ifc->name) < 0) 
						{
							fprintf(stderr, "ERROR: %s device's datalink is not supported!\n", dev);
							goto exit_failure;
						}			
						
						device_set_name(ifc->name);

						if (device_del_promisc() < 0)
							goto exit_failure;
						
						if (device_get_mac_address() < 0)
							goto exit_failure;
						
						if (device_get_inet4_addresses() < 0)
							goto exit_failure;
					
						device_info_print();	
						goto exit_success;
					}
					else 
					{
						fprintf(stderr, "ERROR: %s device is not supported!\n", dev);
						goto exit_failure;
					}
				}
				break;
				
			case (DEVICE_AUTO):
				if ((dev = device_get_default()) == NULL)
					goto exit_failure;
				
				if (strncmp(dev, ifc->name, IF_NAMESIZE) == 0)
				{
					if (device_check_datalink(dev) < 0)
					{
						fprintf(stderr, "ERROR: %s device's datalink is not supported!\n", dev);
						goto exit_failure;
					}
			
					device_set_name(dev);
					
					if (device_del_promisc() < 0)
						goto exit_failure;
					
					if (device_get_mac_address() < 0)
						goto exit_failure;
				
					if (device_get_inet4_addresses() < 0)
						goto exit_failure;
			
					device_info_print();	
					goto exit_success;
				}
				break;
				
			case (DEVICE_LIST):
				if ((ifc->flags & PCAP_IF_LOOPBACK) == 0)
				{
					if (device_check_datalink(ifc->name) < 0)
						break;
					
					device_set_name(ifc->name);
					
					if (device_del_promisc() < 0)
						goto exit_failure;
					
					if (device_get_mac_address() < 0)
						goto exit_failure;
					
					if (device_get_inet4_addresses() < 0)
						goto exit_failure;
			
					device_info_print();
				}
				break;
		}
	}

	if (op_dev == DEVICE_MANUAL)
	{
		if (strncmp(dev, device, IF_NAMESIZE) != 0)
		{
			fprintf(stderr, "ERROR: %s device not found!\n", dev);
			goto exit_failure;
		}
	}
	
	pcap_freealldevs(ifc);
	return (0);
	
	exit_failure:
		pcap_freealldevs(ifc);
		return (-1);

	exit_success:
		pcap_freealldevs(ifc);
		return (0);
}

/*
 * Requests default network interface
 */
static char *
device_get_default()
{
	char *dev, errbuf[PCAP_ERRBUF_SIZE];
	
	if ((dev = pcap_lookupdev(errbuf)) == NULL)
	{
		ERROR(errbuf);
		return (NULL);
	}
	
	return (dev);
}

/*
 * Sets the network interface
 */
static void
device_set_name(char *name)
{

	memset(device, '\0',  IF_NAMESIZE);
	strncpy(device, name, IF_NAMESIZE);
}

/*
 * Checks for the datalink, accepting EN10MB only
 * (and so only ethernet and wireless devices)
 */
static int
device_check_datalink(char *dev)
{
	pcap_t *pcap;
	char errbuf[PCAP_ERRBUF_SIZE];
	int datalink;	

	if ((pcap = pcap_open_live(dev, BUFSIZ, 0, 0, errbuf)) == NULL)
	{
		ERROR(errbuf);
		return (-1);
	}
	
	if ((datalink = pcap_datalink(pcap)) < 0)
	{
		ERROR(pcap_geterr(pcap));
		return (-1);
	}
	
	switch (datalink)
	{
		case (DLT_EN10MB):
			/* 
			 * sets network interface offset.
			 * 14 is ethernet packet len  
			 */
			device_offset = 14;
	
			pcap_close(pcap);
			return (0);
	}
	
	pcap_close(pcap);
	return (-1);
}

/*
 * Putting down the promiscue flag if found set in network interface
 */
static int
device_del_promisc()
{	
	int sd;
	struct ifreq ifr;
	
	if ((sd = socket(AF_INET, SOCK_STREAM, 0)) < 0)
	{
		ERROR(strerror(errno));
		return (-1);
	}
	
	strncpy(ifr.ifr_name, device, sizeof(char) * IF_NAMESIZE);
	ifr.ifr_name[sizeof(char) * strlen(ifr.ifr_name)] = '\0';
	
	if (ioctl(sd, SIOCGIFFLAGS, &ifr) < 0)
	{
		ERROR(strerror(errno));
		goto exit_failure;
	}

	if (ifr.ifr_flags & IFF_PROMISC)
	{
		/* Remove Promisc flag */
		ifr.ifr_flags &= ~IFF_PROMISC;
		if (ioctl(sd, SIOCSIFFLAGS, &ifr) < 0)
		{
			ERROR(strerror(errno));
			goto exit_failure;
		}
	}

	close(sd);
	return (0);
	
	exit_failure:
		close(sd);
		return (-1);
}


/*
 * Reading network interfaces MAC address 
 */
static int
device_get_mac_address()
{
	register int i;
	eth_t *if_eth;
	eth_addr_t mac;
	
	if ((if_eth = eth_open(device)) == NULL)
	{
		ERROR(strerror(errno));
		return (-1);
	}
	
	if (eth_get(if_eth, &mac) < 0)
	{
		ERROR(strerror(errno));
		
		eth_close(if_eth);
		return (-1);
	}
	
	for (i = 0; i < ETHER_ADDR_LEN; i++)
		device_mac.octet[i] = mac.data[i];

	eth_close(if_eth);
	return (0);
}

/*
 * Checks and saves IPv4 and netmask network adresses.
 */
static int 
device_get_inet4_addresses(void)
{
        bpf_u_int32 network;
        char errbuf_lnet[LIBNET_ERRBUF_SIZE], errbuf_pcap[PCAP_ERRBUF_SIZE];
        libnet_t *lnet;

        if ((lnet = libnet_init(LIBNET_LINK, device, errbuf_lnet)) == NULL)
        {
                ERROR(errbuf_lnet);
                return (-1);
        }

        device_inet4.s_addr = libnet_get_ipaddr4(lnet);

        libnet_destroy(lnet);

        /* Get netmask */
        if((pcap_lookupnet(device, &network, &device_netmask.s_addr, errbuf_pcap)) < 0)
        {
                ERROR(errbuf_pcap);
                return (-1);
        }

        return (0);
}

/*
 * Prints network interfaces dates
 */
static void 
device_info_print()
{
	
	printf("  - Device: (%s)  Datalink: Ethernet  Mac: %s\n", device, ether_ntoa(&device_mac));
        printf("    Inet4: %15s", inet_ntoa(device_inet4));
        printf("  Netmask: %15s\n", inet_ntoa(device_netmask));
} 
	 
/*
 * ARP PING handler:
 */

/*
 * Set ping response timeout
 */
static int 
arp_ping_set_timeout(char *timeout)
{
	/* Read arp_ping_recv_packet() for ERROR info */
#ifdef LINUX
	fprintf(stderr, "ERROR: ping timeout not supported on Linux platform.\n");
	fprintf(stderr, "       You can to read \"man 3 select\" BUGS section.\n");

	return (-1);
#else
	ping_timeout = atoi(timeout);

	return (0);
#endif
}

/*
 * Ping to single host, sends a sole Arp request
 * and waits a sole Arp reply
 */
static int
arp_ping_host(char *arg)
{
	struct ether_addr dst_mac;
	struct in_addr dst_ip;
	pcap_t *pcap;

	memcpy(&dst_mac, ether_aton(MAC_ADDR_BROADCAST), sizeof(dst_mac));
	
	if (inet_aton(arg, &dst_ip) == 0)
	{
		fprintf(stderr, "ERROR: %s inet4 address malformed!\n", arg);
		return (-1);
	}
	
	if ((pcap = arp_ping_open_packet()) == NULL)
		return (-1);
	
	printf("  - Arp Ping response timeout: %d ms.\n", ping_timeout); 
	printf("  - Arp Ping to: %s\n", arg);
	if (arp_ping_send_packet(PING_HOST, &dst_mac, &dst_ip) < 0)
		goto exit_failure;
	
	if (arp_ping_recv_packet(PING_HOST, pcap, (unsigned char *) arg) < 0)
		goto exit_failure;
	
	arp_ping_close_packet(pcap);
	return (0);
	
	exit_failure:
		arp_ping_close_packet(pcap);
		return (-1);
}

/*
 * It sends out N Arp requests and waits for N Arp
 * reply, where N is the number of host
 */
static int
arp_ping_broadcast()
{
	int n_hosts, netmask_c, netmask_b, netmask_a;
	int inet4_c, inet4_b, inet4_a;
	register int c, b, a;
	char host[INET4_ADDR_LEN];
	struct ether_addr dst_mac;
	struct in_addr dst_ip;
	pcap_t *pcap;
	
	/* Number hosts */
	n_hosts	= ntohl(~device_netmask.s_addr);
	
#if BYTE_ORDER == LITTLE_ENDIAN
        /* Netmask Classes */
        netmask_c = device_netmask.s_addr >> 24 & 0xff,
        netmask_b = device_netmask.s_addr >> 16 & 0xff,
        netmask_a = device_netmask.s_addr >> 8 & 0xff;

        /* Broadcast Classes */
        inet4_c = device_inet4.s_addr >> 16 & 0xff,
        inet4_b = device_inet4.s_addr >> 8 & 0xff,
        inet4_a = device_inet4.s_addr & 0xff;
#else
        /* Netmask Classes */
        netmask_c = device_netmask.s_addr & 0xff,
        netmask_b = device_netmask.s_addr >> 8 & 0xff,
        netmask_a = device_netmask.s_addr >> 16 & 0xff,

        /* Broadcast Classes */
        inet4_c = device_inet4.s_addr >> 8 & 0xff,
        inet4_b = device_inet4.s_addr >> 16 & 0xff,
        inet4_a = device_inet4.s_addr >> 24 & 0xff;
#endif

        printf("  - Arp Ping response timeout: %d ms.\n", ping_timeout); 
        printf("  - Arp ping to Broadcast with Netmask: %s  ", inet_ntoa(device_netmask));
	
	printf("Class: \"%c\"  Possible Hosts: %d\n", 
		n_hosts <= NETMASK_CLASS_C ? 'C' :
		n_hosts <= NETMASK_CLASS_B ? 'B' :
		n_hosts <= NETMASK_CLASS_A ? 'A' : '?' , n_hosts);
	
	printf("  - LAN's active hosts:\n");
	
	if (n_hosts <= NETMASK_CLASS_C)
	{
		for (c = netmask_c; c < NETMASK_CLASS_C; c++)
		{
			if ((pcap = arp_ping_open_packet()) == NULL)
				return (-1);

			memcpy(&dst_mac, ether_aton(MAC_ADDR_BROADCAST), sizeof(dst_mac));
			
			sprintf(host, "%d.%d.%d.%d", inet4_a, inet4_b, inet4_c, c);	
			if (inet_aton(host, &dst_ip) == 0)
			{
				fprintf(stderr, "ERROR: %s inet4 address malformed!\n", host);
				goto exit_failure;
			}
			
			if (arp_ping_send_packet(PING_BROADCAST, &dst_mac, &dst_ip) < 0)
				goto exit_failure;
			   
			if (arp_ping_recv_packet(PING_BROADCAST, pcap, (unsigned char *) host) < 0)
				goto exit_failure;
		
			arp_ping_close_packet(pcap);
		}
	}
	else if (n_hosts <= NETMASK_CLASS_B)
	{
		for (b = netmask_b; b < NETMASK_CLASS_C; b++)
		{
			for (c = netmask_c; c < NETMASK_CLASS_C; c++)
			{
				if ((pcap = arp_ping_open_packet()) == NULL)
					return (-1);

				memcpy(&dst_mac, ether_aton(MAC_ADDR_BROADCAST), sizeof(dst_mac));
			
				sprintf(host, "%d.%d.%d.%d", inet4_a, inet4_b, b, c);	
				if (inet_aton(host, &dst_ip) == 0) 
				{
					fprintf(stderr, "ERROR: %s inet4 address malformed!\n", host);
					goto exit_failure;
				}
				
				if (arp_ping_send_packet(PING_BROADCAST, &dst_mac, &dst_ip) < 0)
					goto exit_failure;
			   
				if (arp_ping_recv_packet(PING_BROADCAST, pcap, (unsigned char *) host) < 0)
					goto exit_failure;
				
				arp_ping_close_packet(pcap);
			}
		}
	}
	else if (n_hosts <= NETMASK_CLASS_A)
	{
		for (a = netmask_a; a < NETMASK_CLASS_C; a++)
		{
			for (b = netmask_b; b < NETMASK_CLASS_C; b++)
			{
				for (c = netmask_c; c < NETMASK_CLASS_C; c++)
				{
					if ((pcap = arp_ping_open_packet()) == NULL)
						return (-1);

					memcpy(&dst_mac, ether_aton(MAC_ADDR_BROADCAST), sizeof(dst_mac));
			
					sprintf(host, "%d.%d.%d.%d", inet4_a, a, b, c);	
					if (inet_aton(host, &dst_ip) == 0)
					{
						fprintf(stderr, "ERROR: %s inet4 address malformed!\n", host);
						goto exit_failure;
					}
				
					if (arp_ping_send_packet(PING_BROADCAST, &dst_mac, &dst_ip) < 0)
						goto exit_failure;
			   
					if (arp_ping_recv_packet(PING_BROADCAST, pcap, (unsigned char *) host) < 0)
						goto exit_failure;
				
					arp_ping_close_packet(pcap);
				}
			}
		}
	}
	/* Network Class unknown */
	else
	{
		fprintf(stderr, "ERROR: netmask is not correct!\n");
		return (-1);
	}
	
	return (0);	
	
	exit_failure:
		arp_ping_close_packet(pcap);
		return (-1);
}

/*
 * Open pcap file descriptor
 */
static pcap_t *
arp_ping_open_packet()
{
	char *filter = "arp", errbuf[PCAP_ERRBUF_SIZE];
	struct bpf_program compiled_filter;
	pcap_t *pcap;

	/* 
	 * Linux doesn't support timeout on pcap_open_live(),
	 * i opted to use poll() function on arp_ping_recv_packet()
	 */
#ifdef LINUX
	if ((pcap = pcap_open_live(device, BUFSIZ, 0, 0, errbuf)) == NULL)
#else
	if ((pcap = pcap_open_live(device, BUFSIZ, 0, ping_timeout, errbuf)) == NULL)
#endif
	{
		ERROR(errbuf);
		return (NULL);
	}

	if (pcap_compile(pcap, &compiled_filter, filter, 0, device_netmask.s_addr) < 0)
	{	
		ERROR(pcap_geterr(pcap));	
		goto exit_failure;
	}
	
	if (pcap_setfilter(pcap, &compiled_filter) < 0)
	{
		ERROR(pcap_geterr(pcap));
		goto exit_failure;
	}	
	
	pcap_freecode(&compiled_filter);

	/* Input Direction */
	if (pcap_setdirection(pcap, PCAP_D_IN) < 0)
	{
		ERROR(pcap_geterr(pcap));
		goto exit_failure;
	}
	
	return (pcap);
	
	exit_failure:
		arp_ping_close_packet(pcap);
		return (NULL);
}

/*
 * Close pcap file descriptor
 */
static void
arp_ping_close_packet(pcap_t *pcap)
{

	pcap_close(pcap);
}

/*
 * It receives the packet using a 500/user defined ms timeout,
 * if it arrives it starts reading
 */
static int
arp_ping_recv_packet(unsigned int type, pcap_t *pcap, unsigned char *arg)
{
	/* 
	 * Linux doesn't support timeout on pcap_open_live(),
	 * i opted to use poll() but i don't to resolve this problem
         * (see: man 3 select - BUGS section)
         */
#ifdef LINUX
	struct pollfd fds;

	 /* This variable is used by arp_ping_read_packet() */
        ping_type = type;

	fds.fd = pcap_fileno(pcap);
	fds.events = POLLIN;
	fds.revents = 0x0;

	switch (poll(&fds, 1, ping_timeout))
	{
		case (-1):
			ERROR(strerror(errno));
			return (-1);

		case (0):
			switch (type)
			{
				case (PING_HOST):
					fprintf(stderr, "ERROR: %s host is down!\n", arg);
					return (-1);

				case (PING_BROADCAST):
					return (0);
			}
	
		default:
			if (pcap_loop(pcap, 1, arp_ping_read_packet, arg) < 0)
                        {
                        	ERROR(pcap_geterr(pcap));
                               	return (-1);
                     	}
	}
#else	
	 /* This variable is used by arp_ping_read_packet() */
        ping_type = type;

	if (pcap_dispatch(pcap, 1, arp_ping_read_packet, arg) < 0)
	{
               	ERROR(pcap_geterr(pcap));
               	return (-1);
       	}
#endif
	/* ping_check_reply shows the value signed by arp_ping_read_packet() */
	if (type == PING_HOST && ping_check_reply == PING_REPLY_FALSE)
	{
		fprintf(stderr, "ERROR: %s host is down!\n", arg);
		return (-1);
	}
	
	return (0);
}

/*
 * Reads the inbound packet from Arp ping,
 * accepts only the Arp reply from the trusted host of request
 */
static void
arp_ping_read_packet(unsigned char *arg, const struct pcap_pkthdr *header, const unsigned char *packet)
{
	register int i;
	struct ether_addr src_mac;
	struct in_addr src_ip;
	char c_src_ip[INET4_ADDR_LEN];

	arp_packet = (struct arp_header *) (packet + device_offset);
	
	if (ntohs(arp_packet->header.ar_op) == ARP_OP_REPLY)
	{
		sprintf(c_src_ip, "%d.%d.%d.%d", 
			arp_packet->addresses.ar_spa[0], arp_packet->addresses.ar_spa[1],
			arp_packet->addresses.ar_spa[2], arp_packet->addresses.ar_spa[3]);
		inet_aton(c_src_ip, &src_ip);
		
		if (strcmp((char *) arg, inet_ntoa(src_ip)) == 0)
		{
			for (i = 0; i < ETHER_ADDR_LEN; i++)
				src_mac.octet[i] = arp_packet->addresses.ar_sha[i];

			/* This variable is assigned by arp_ping_recv_packets() */			
			switch (ping_type)
			{
				case (PING_HOST):
					printf("  <- Arp reply %s is-at (%s)\n", 
					      	inet_ntoa(src_ip), ether_ntoa(&src_mac));
					break;
				
				case (PING_BROADCAST):
					printf("  %d) Inet4: %15s  ->  Mac: %17s\n", 
						ping_broadcast_count, 
						inet_ntoa(src_ip), ether_ntoa(&src_mac));
						   
					ping_broadcast_count++;
					break;
			}
		
			/* Successfully received Arp reply from the host */
			ping_check_reply = PING_REPLY_TRUE;
		}
		else
		{	
			/* Received Arp reply doesn't come from host */
			ping_check_reply = PING_REPLY_FALSE;
		}
	}
	else
	{
		/* This packet is not Arp reply */
		ping_check_reply = PING_REPLY_FALSE;
	}			
	
	/* 
	 * ping_check_reply will checks later from 
	 * arp_ping_recv_packets()
	 */	
}

/*
 * Sends Arp request packet for Arp ping
 */
static int
arp_ping_send_packet(unsigned int type, struct ether_addr *dst_mac, struct in_addr *dst_ip)
{
	libnet_t *lnet;
	char errbuf[LIBNET_ERRBUF_SIZE];

	if ((lnet = libnet_init(LIBNET_LINK, device, errbuf)) == NULL)
	{
		ERROR(errbuf);
		return (-1);
   	}
	
	if (libnet_autobuild_arp(ARPOP_REQUEST, 
				 (u_int8_t *) &device_mac,	/* Source Mac */
				 (u_int8_t *) &device_inet4,	/* Source InetV4 */
				 (u_int8_t *) dst_mac,		/* Destination Mac */
				 (u_int8_t *) dst_ip,		/* Destination InetV4 */
				 lnet ) < 0)
	{
		ERROR(libnet_geterror(lnet));
		goto exit_failure;
	}
	
	if (libnet_autobuild_ethernet((u_int8_t *) dst_mac, ETHERTYPE_ARP, lnet) < 0)
	{
		ERROR(libnet_geterror(lnet));
		goto exit_failure;
	}
	
	/* Sends (Ethernet + Arp) Packet */
	if (libnet_write(lnet) < 0)
	{
		ERROR(libnet_geterror(lnet));
		goto exit_failure;
	}
	
	if (type == PING_HOST)
	{
		printf("  -> Arp who-has %s (%s)", 
		       inet_ntoa(*dst_ip), ether_ntoa(dst_mac));
		printf(" tell %s (%s)\n",
		       inet_ntoa(device_inet4), ether_ntoa(&device_mac));
	}
		
	libnet_destroy(lnet);
	return (0);
	
	exit_failure:
		libnet_destroy(lnet);
		return (-1);
}

/*
 * ARP SNIFF Packets handler:
 */

/*
 * Process only inbound packets:
 *     - Arp reply
 *     - Arp request
 */
static int
arp_sniff_packets()
{
	pcap_t *pcap;
	
	if ((pcap = arp_sniff_open_packets()) == NULL)
		return (-1);
		
	printf("  - Sniffing Arp packets:\n");

	/* Used by arp_sniff_signal */ 	
	pcap_sniff = pcap;
	/* Ctrl + C */
	signal(SIGINT, arp_sniff_signal);
		
	if (pcap_loop(pcap, -1, arp_sniff_read_packets, NULL) < 0)
	{
		ERROR(pcap_geterr(pcap));
		goto exit_failure;
	}
	
	arp_sniff_close_packets(pcap);
	return (0);
	
	exit_failure:
		arp_sniff_close_packets(pcap);
		return (-1);

}

/*
 * If SIGINT signal (Ctrl + C) arrives,
 * it goes back to the passively sniffed
 * packet's stats and then it exits
 */
static void
arp_sniff_signal(int sig)
{
	struct pcap_stat stats;
	
	if (pcap_stats(pcap_sniff, &stats) < 0)
	{
		ERROR(pcap_geterr(pcap_sniff));
		exit(EXIT_FAILURE);
	}

	printf("\r  - Packets stats:\n");
	printf("    Received: %d\n",		     stats.ps_recv);
	printf("    Received \"Arp Request\": %d\n", arp_sniff_request_count);
	printf("    Received \"Arp Reply\": %d\n",   arp_sniff_reply_count);
	printf("    Dropped: %d\n",		     stats.ps_drop);
	printf("    Dropped by interface: %d\n",     stats.ps_ifdrop);
		   
	exit(EXIT_SUCCESS);
}

/*
 * Open pcap file descriptor
 */
static pcap_t *
arp_sniff_open_packets()
{
	char *filter = "arp", errbuf[PCAP_ERRBUF_SIZE];
	struct bpf_program compiled_filter;
	pcap_t *pcap;

	if ((pcap = pcap_open_live(device, BUFSIZ, 0, 0, errbuf)) == NULL)
	{
		ERROR(errbuf);
		return (NULL);
	}

	if (pcap_setnonblock(pcap, 1, errbuf) < 0)
	{
		ERROR(pcap_geterr(pcap));
		goto exit_failure;
	}

	if (pcap_compile(pcap, &compiled_filter, filter, 0, device_netmask.s_addr) < 0)
	{		
		ERROR(pcap_geterr(pcap));
		goto exit_failure;
	}
	
	if (pcap_setfilter(pcap, &compiled_filter) < 0)
	{
		ERROR(pcap_geterr(pcap));
		goto exit_failure;
	}	
	
	pcap_freecode(&compiled_filter);

	return (pcap);
	
	exit_failure:
		arp_sniff_close_packets(pcap);
		return (NULL);
}	

/*
 * Close pcap file descriptor
 */
static void
arp_sniff_close_packets(pcap_t *pcap)
{

	pcap_close(pcap);
}

/*
 * Reads inbound packets:
 *     - Arp reply
 *     - Arp request
 */
static void
arp_sniff_read_packets(unsigned char *arg, const struct pcap_pkthdr *header, const unsigned char *packet)
{
	register int i;
	char c_src_ip[INET4_ADDR_LEN], c_dst_ip[INET4_ADDR_LEN];
	struct ether_addr src_mac, dst_mac;
	struct in_addr src_ip, dst_ip;

	arp_packet = (struct arp_header *) (packet + device_offset);

	if (ntohs(arp_packet->header.ar_op) == ARP_OP_REPLY)
	{
		sprintf(c_src_ip, "%d.%d.%d.%d", 
			arp_packet->addresses.ar_spa[0], arp_packet->addresses.ar_spa[1],
			arp_packet->addresses.ar_spa[2], arp_packet->addresses.ar_spa[3]);
		inet_aton(c_src_ip, &src_ip);
		
		for (i = 0; i < ETHER_ADDR_LEN; i++)
			src_mac.octet[i] = arp_packet->addresses.ar_sha[i];
	
		/* Count Arp reply */
		arp_sniff_reply_count += 1;
		
		
		printf("  <- Arp reply %s is-at (%s)\n", 
			   inet_ntoa(src_ip), ether_ntoa(&src_mac));
	}
	else if (ntohs(arp_packet->header.ar_op) == ARP_OP_REQUEST)
	{
		sprintf(c_src_ip, "%d.%d.%d.%d", 
			arp_packet->addresses.ar_spa[0], arp_packet->addresses.ar_spa[1],
			arp_packet->addresses.ar_spa[2], arp_packet->addresses.ar_spa[3]);
		inet_aton(c_src_ip, &src_ip);
		
		for (i = 0; i < ETHER_ADDR_LEN; i++)
			src_mac.octet[i] = arp_packet->addresses.ar_sha[i];
	
		sprintf(c_dst_ip, "%d.%d.%d.%d", 
			arp_packet->addresses.ar_tpa[0], arp_packet->addresses.ar_tpa[1],
			arp_packet->addresses.ar_tpa[2], arp_packet->addresses.ar_tpa[3]);
		inet_aton(c_dst_ip, &dst_ip);
	
		for (i = 0; i < ETHER_ADDR_LEN; i++)
			dst_mac.octet[i] = arp_packet->addresses.ar_tha[i];
	
		/* Count Arp request */
		arp_sniff_request_count += 1;
	
		printf("  -> Arp who-has %s (%s)", 
			   inet_ntoa(dst_ip), ether_ntoa(&dst_mac));
		printf(" tell %s (%s)\n",
			   inet_ntoa(src_ip), ether_ntoa(&src_mac));
	}
}

/*
 * ARP CACHE handler:
 */

/*
 * It handles the Arp cache's functionally:
 *     - Add 
 *     - Delete 
 *     - List Arp cache
 */
static int
arp_cache_manager(unsigned int op_ac, char *entry)
{
	arp_t *arp;

	if ((arp = arp_open()) == NULL)
	{
		ERROR(strerror(errno));
		return (-1);
	}
		
	switch (op_ac)
	{
		case (ARP_CACHE_ADD):
			if (arp_cache_add(ARP_CACHE, arp, entry) < 0)
				goto exit_failure;		
			break;
			
		case (ARP_CACHE_DEL):
			if (arp_cache_del(ARP_CACHE, arp, entry) < 0)
				goto exit_failure;
			break;
			
		case (ARP_CACHE_LIST):
			if (arp_cache_list(arp) < 0)
				goto exit_failure;
			break;
	}
	
	arp_close(arp);
	return (0);
	
	exit_failure:
		arp_close(arp);
		return (-1);
}

/*
 * Adds Arp cache entry
 */
static int
arp_cache_add(unsigned int type, arp_t *arp, char *arg)
{
	register int i;
	char *c_mac, *c_ip;
	struct arp_entry entry;
	struct ether_addr mac;
	struct in_addr ip;

	if ((c_mac = malloc(sizeof(char) * MAC_ADDR_LEN + 1)) == NULL)
	{
		ERROR(strerror(errno));
		return (-1);
	}
	
	if ((c_ip = malloc(sizeof(char) * INET4_ADDR_LEN + 1)) == NULL)
	{
		ERROR(strerror(errno));
		return (-1);
	}
	
	sscanf(arg, "%15s %17s", c_ip, c_mac);
	
	if (inet_aton(c_ip, &ip) == 0)
	{
		fprintf(stderr, "ERROR: %s inet4 address malformed!\n", c_ip);
		return (-1);
	}
	
	if (ether_aton(c_mac) == NULL)
	{
		fprintf(stderr, "ERROR: %s mac address malformed!\n", c_mac);
		return (-1);
	}
	
	memcpy(&mac, ether_aton(c_mac), sizeof(mac));
	
	entry.arp_pa.addr_type = ADDR_TYPE_IP;
	entry.arp_ha.addr_type = ADDR_TYPE_ETH;
				
	for (i = 0; i < ETHER_ADDR_LEN; i++)
		entry.arp_ha.addr_eth.data[i] = mac.octet[i];

	entry.arp_pa.addr_ip = ip.s_addr;

	if (arp_add(arp, &entry) < 0)
	{
		ERROR(strerror(errno));
		return (-1);
	}
	
	/*
	 * These outputs depends by asked functionally
	 */	
	switch (type)
	{
		case (ARP_CACHE):
			printf("  Arp Cache:\n");
			printf("  - Inet4: %15s  ->  Mac: %17s entry added!\n", 
				   inet_ntoa(ip), ether_ntoa(&mac));
			break;
		
		case (SARPI):
			printf("    To:    Inet4: %15s  ->  Mac: %17s entry\n",
				   inet_ntoa(ip), ether_ntoa(&mac));
			break;
			
		case (DARPI):
			printf("  - DARPI Realtime adds entry in Arp Cache:\n");
			printf("    Inet4: %15s  ->  Mac: %17s entry\n",
				   inet_ntoa(ip), ether_ntoa(&mac));
			break;
	}
	
	free(c_mac);
	free(c_ip);
	
	return (0);
}

/*
 * Search entry in Arp cache, if is found it will delete 
 */
static int
arp_cache_del(unsigned int type, arp_t *arp, char *arg)
{
	unsigned int i;
	struct arp_entry entry;

	/* Read Arp cache */
	if (arp_loop(arp, arp_cache_list_create, NULL) < 0)
		return (-1);
	
	/* Search InetV4 or Mac in Arp Cache */
	TAILQ_FOREACH(arp_cache_pos, &arp_cache_head, entries)
	{
		if (strcmp(arg, inet_ntoa(arp_cache_pos->ip)) == 0 || 
		    strcmp(arg, ether_ntoa(&arp_cache_pos->mac)) == 0)
		{
			entry.arp_pa.addr_type = ADDR_TYPE_IP;
				
			for (i = 0; i < ETHER_ADDR_LEN; i++)
				entry.arp_ha.addr_eth.data[i] = arp_cache_pos->mac.octet[i];
			
			entry.arp_pa.addr_ip = arp_cache_pos->ip.s_addr;
	
			if (arp_delete(arp, &entry) < 0)
			{
				ERROR(strerror(errno));

				arp_cache_list_destroy();
				return (-1);
			}
		
			/*
			 * Depend by asked functionally
			 */	
			switch (type)
			{
				case (ARP_CACHE):
					printf("  Arp Cache:\n");
					printf("  - Inet4: %15s  ->  Mac: %17s entry deleted!\n", 
						   inet_ntoa(arp_cache_pos->ip), ether_ntoa(&arp_cache_pos->mac));
					break;
					
				case (DARPI):
					printf("  - DARPI Realtime deletes entry in Arp Cache:\n");
					printf("    Inet4: %15s  ->  Mac: %17s entry\n",
						   inet_ntoa(arp_cache_pos->ip), ether_ntoa(&arp_cache_pos->mac));
					break;
			}
			
			arp_cache_list_destroy();
			return (0);
		}
	}
	
	fprintf(stderr, "ERROR: %s entry not found!\n", arg);
	
	arp_cache_list_destroy();
	return (-1);
}

/*
 * Prints Arp cache
 */
static int
arp_cache_list(arp_t *arp)
{

	printf("  - Arp Cache list:\n");
	if (arp_loop(arp, arp_cache_list_create, NULL) < 0)
		return (-1);
			
	arp_cache_list_print();
	arp_cache_list_destroy();
	
	return (0);
}

/*
 * Create Arp cache tail using one entry
 * or it adds simply a new entry
 */
static int
arp_cache_list_create(const struct arp_entry *entry, void *arg)
{
	register int i;
	struct ether_addr mac;
	struct in_addr ip;

	for (i = 0; i < ETHER_ADDR_LEN; i++)
		mac.octet[i] = entry->arp_ha.addr_eth.data[i];
	
	ip.s_addr = entry->arp_pa.addr_ip;
	
	if (arp_cache_begin == NULL) 
	{
		TAILQ_INIT(&arp_cache_head);
	
		if ((arp_cache_begin = malloc(sizeof(struct arp_cache))) == NULL)
		{
			ERROR(strerror(errno));
			return (-1);
		}
		memcpy(&arp_cache_begin->mac, &mac, sizeof(mac));
		memcpy(&arp_cache_begin->ip, &ip, sizeof(ip));
	
		TAILQ_INSERT_HEAD(&arp_cache_head, arp_cache_begin, entries);
	}
	else
	{
		if ((arp_cache_next = malloc(sizeof(struct arp_cache))) == NULL)
		{
			ERROR(strerror(errno));
			return (-1);
		}
		memcpy(&arp_cache_next->mac, &mac, sizeof(mac));
		memcpy(&arp_cache_next->ip, &ip, sizeof(ip));

		TAILQ_INSERT_TAIL(&arp_cache_head, arp_cache_next, entries);
	}
	
	return (0);
}

/*
 * Prints Arp cache tail
 */
static void
arp_cache_list_print(void)
{
	unsigned int i = 1;

	TAILQ_FOREACH(arp_cache_pos, &arp_cache_head, entries)
	{
		printf("  %d) Inet4: %15s  ->  Mac: %17s\n", 
			i, inet_ntoa(arp_cache_pos->ip), ether_ntoa(&arp_cache_pos->mac));
		
		i++;
	}
	
	return;
}

/*
 * Destroy Arp cache tail
 */
static void
arp_cache_list_destroy(void)
{

	while (TAILQ_EMPTY(&arp_cache_head) == 0)
	{
		arp_cache_pos = TAILQ_FIRST(&arp_cache_head);
		
		TAILQ_REMOVE(&arp_cache_head, arp_cache_pos, entries);
		free(arp_cache_pos);
     	}
}

/*
 * SARPI handler:
 */

/*
 * Sets Arp cache timeout for automatic update
 */
static void
sarpi_set_timeout(char *timeout)
{

	sarpi_timeout = atoi(timeout);
}

/*
 * Handles SARPI through two processes:
 *     - Child:	
 *     	   	Update automatically the Arp cache
 *     - Parent:	
 *		Works in real time, in other words
 *		it listens to the inbound packets	
 */
static int
sarpi_manager()
{
	time_t time_cur;
	struct tm tm_cur;
	arp_t *arp;
	
	/* Localtime */
	time_cur = time(NULL);
    	tm_cur = *(struct tm *) localtime(&time_cur);
	printf("  - SARPI Start at %d:%d:%d ...\n", tm_cur.tm_hour, tm_cur.tm_min, tm_cur.tm_sec);
	
	if ((arp = sarpi_arp_cache_open()) == NULL)
		return (-1);
	
	/* Arp Cache entries protect */
	sarpi_arp_cache_list();
	
	switch ((sarpi_pid_child = fork()))
	{
		case (-1):
			ERROR(strerror(errno));
			return (-1);
		/* Child */
		case (0):
			/* Arp Cache refresh with timeout */
			if (sarpi_arp_cache_refresh(arp) < 0)
				goto exit_failure;
			break;

		default:
			break;
	}
	
	/* Ctrl + C */
	signal(SIGINT, sarpi_manager_signal);
	signal(SIGTERM, sarpi_manager_signal);
	signal(SIGKILL, sarpi_manager_signal);
	
	/* 
	 * Parent Task
	 * SARPI Realtime
	 */
	if (sarpi_realtime(arp) < 0)
		goto exit_failure;
	
	sarpi_arp_cache_close(arp);
	return (0);
		
	exit_failure:
		sarpi_arp_cache_close(arp);
		return (-1);
}

/*
 * If SIGINT signal (Ctrl + C) or SIGTERM/SIGKILL arrives,
 * it kills child process and then it exits
 */
static void
sarpi_manager_signal(int sig)
{

	kill(sarpi_pid_child, SIGTERM);
	
	exit(EXIT_SUCCESS);
}

/*
 * Open or create Arp cache tail
 */
static arp_t * 
sarpi_arp_cache_open()
{
	arp_t *arp;
	
	if ((arp = arp_open()) == NULL)
	{
		ERROR(strerror(errno));
		return (NULL);
	}
	
	if (arp_loop(arp, arp_cache_list_create, NULL) < 0)
	{
		arp_close(arp);
		return (NULL);
	}
			
	return (arp);
}

/*
 * Destroy and close Arp cache tail
 */
static void
sarpi_arp_cache_close(arp_t *arp)
{

	arp_cache_list_destroy();
	arp_close(arp);
}

/*
 * Prints Arp cache tail which contains
 * the entries protected by SARPI
 */
static void
sarpi_arp_cache_list()
{

	printf("  - SARPI protects these Arp Cache's entries:\n");
	arp_cache_list_print();
}

/*
 * During every timeout it updates
 * static entries in kernel's Arp cache
 */
static int
sarpi_arp_cache_refresh(arp_t *arp)
{
	int entry_len;
	char *entry;

	printf("  - SARPI Arp Cache refresh timeout: %d %s\n", 
	       sarpi_timeout, sarpi_timeout == 1 ? "minut." : "minuts.");
	
	for (;;)
	{
		/* timeout * 60 sec = N minuts */
		sleep(sarpi_timeout * 60);
		
		printf("  - SARPI refresh these Arp Cache entries:\n");
		
		TAILQ_FOREACH(arp_cache_pos, &arp_cache_head, entries)
		{
			entry_len = strlen(inet_ntoa(arp_cache_pos->ip)) + strlen(ether_ntoa(&arp_cache_pos->mac)) + 2;
			
			if ((entry = malloc(sizeof(char) * entry_len)) == NULL)
			{
				ERROR(strerror(errno));
				return (-1);
			}
			sprintf(entry, "%s %s", inet_ntoa(arp_cache_pos->ip), ether_ntoa(&arp_cache_pos->mac));
			
			/* Refresh Arp Cache entry */
			if (arp_cache_add(SARPI, arp, entry) < 0)
			{
				free(entry);
				return (-1);
			}
		
			free(entry);
		}
	}
	
	return (0);
}

/*
 * SARPI Realtime,
 * process Arp reply inbound packets
 */
static int
sarpi_realtime(arp_t *arp)
{
	pcap_t *pcap;

	if ((pcap = sarpi_realtime_open_packets()) == NULL)
		return (-1);
	
	printf("  - SARPI Realtime Protect actived\n");
	
	/* It used by sarpi_realtime_read_packets() */
	sarpi_arp = arp;
	
	if (pcap_loop(pcap, -1, sarpi_realtime_read_packets, NULL) < 0)
	{
		ERROR(pcap_geterr(pcap));
		
		sarpi_realtime_close_packets(pcap);
		return (-1);
	}
	
	sarpi_realtime_close_packets(pcap);
	return (0);
}

/*
 * Open pcap file descriptor
 */
static pcap_t *
sarpi_realtime_open_packets()
{
	char *filter = "arp", errbuf[PCAP_ERRBUF_SIZE];
	struct bpf_program compiled_filter;
	pcap_t *pcap;

	if ((pcap = pcap_open_live(device, BUFSIZ, 0, 0, errbuf)) == NULL)
	{
		ERROR(errbuf);
		return (NULL);
	}

	if (pcap_setnonblock(pcap, 1, errbuf) < 0)
	{
		ERROR(pcap_geterr(pcap));
		goto exit_failure;
	}

	if (pcap_compile(pcap, &compiled_filter, filter, 0, device_netmask.s_addr) < 0)
	{		
		ERROR(pcap_geterr(pcap));
		goto exit_failure;
	}
	
	if (pcap_setfilter( pcap, &compiled_filter ) < 0)
	{
		ERROR(pcap_geterr(pcap));
		goto exit_failure;
	}	
	
	pcap_freecode(&compiled_filter);
	
	return (pcap);
		
	exit_failure:
		sarpi_realtime_close_packets(pcap);
		return (NULL);
}

/*
 * Close pcap file descriptor
 */
static void
sarpi_realtime_close_packets(pcap_t * pcap)
{

	pcap_close(pcap);
}

/*
 * Reads the Arp reply/request packets.
 * (it checks destionation address)
 * If Arp reply/request's data like addresses IP and MAC match a static entry of
 * Arp cache protected by SARPI, update the entry using the static one
 */
static void
sarpi_realtime_read_packets(unsigned char *arg, const struct pcap_pkthdr *header, const unsigned char *packet)
{
	register int i;
	int entry_len;
	char 	*entry, c_src_ip[INET4_ADDR_LEN], c_src_mac[MAC_ADDR_LEN], c_dst_ip[INET4_ADDR_LEN],
		 c_ap_ip[INET4_ADDR_LEN], c_ap_mac[MAC_ADDR_LEN];
	struct ether_addr src_mac;
	struct in_addr src_ip;

	arp_packet = (struct arp_header *) (packet + device_offset);

	if (ntohs(arp_packet->header.ar_op) == ARP_OP_REPLY ||
	    ntohs(arp_packet->header.ar_op) == ARP_OP_REQUEST)
	{
		sprintf(c_src_ip, "%d.%d.%d.%d", 
			arp_packet->addresses.ar_spa[0], arp_packet->addresses.ar_spa[1],
			arp_packet->addresses.ar_spa[2], arp_packet->addresses.ar_spa[3]);
		inet_aton(c_src_ip, &src_ip);
		
		for (i = 0; i < ETHER_ADDR_LEN; i++)
			src_mac.octet[i] = arp_packet->addresses.ar_sha[i];	
		strncpy(c_src_mac, ether_ntoa(&src_mac), sizeof(char) * MAC_ADDR_LEN);
		c_src_mac[sizeof(char) * strlen(c_src_mac)] = '\0';

		sprintf(c_dst_ip, "%d.%d.%d.%d", 
			arp_packet->addresses.ar_tpa[0], arp_packet->addresses.ar_tpa[1],
			arp_packet->addresses.ar_tpa[2], arp_packet->addresses.ar_tpa[3]);
	
		/* Check destionation address */
		if (strcmp(c_dst_ip, inet_ntoa(device_inet4)) == 0)
		{
			/* Search Arp reply/request source in Arp cache entries */
			TAILQ_FOREACH(arp_cache_pos, &arp_cache_head, entries)
			{
				strncpy(c_ap_ip, inet_ntoa(arp_cache_pos->ip), sizeof(char) * INET4_ADDR_LEN);
				c_ap_ip[sizeof(char) * strlen(c_ap_ip)] = '\0';
				
				strncpy(c_ap_mac, ether_ntoa(&arp_cache_pos->mac), sizeof(char) * MAC_ADDR_LEN);
				c_ap_mac[sizeof(char) * strlen(c_ap_mac)] = '\0';

				if (strcmp(c_src_ip , c_ap_ip) == 0 || strcmp(c_src_mac, c_ap_mac) == 0)
				{
					/* Entry Found! */
					switch (ntohs(arp_packet->header.ar_op))
					{
						case (ARP_OP_REPLY):
							printf("  - SARPI ===== ARP REPLY =====\n");
							break;
						case (ARP_OP_REQUEST):
							printf("  - SARPI ===== ARP REQUEST =====\n");
							break;
					}

					printf("  - SARPI RealTime refresh entry in Arp Cache:\n");
					printf("    From:  Inet4: %15s  ->  Mac: %17s entry\n",
					   inet_ntoa(src_ip), ether_ntoa(&src_mac));

					entry_len = strlen(inet_ntoa(arp_cache_pos->ip)) + 
							strlen(ether_ntoa(&arp_cache_pos->mac)) + 2;

					if ((entry = malloc(sizeof(char) * entry_len)) == NULL)
					{
						ERROR(strerror(errno));
						exit(EXIT_FAILURE);
					}

					sprintf(entry, "%15s %17s", inet_ntoa(arp_cache_pos->ip), 
						ether_ntoa(&arp_cache_pos->mac));
				
					/* Refresh Arp Cache entry */
					if (arp_cache_add(SARPI, sarpi_arp, entry) < 0)
					{
						free(entry);
						exit(EXIT_FAILURE);
					} 
		
					free(entry);
				
					return;
				}
			}
		}
	}
}

/*
 * DARPI handler:
 */

/* 
 * Handles DARPI, delete all found entries in the Arp cache to
 * delete some poisoned hosts then it starts realtime execution
 * to reads the packets:
 *     - Arp request
 *     - Arp reply
 */
static int
darpi_manager()
{
	time_t time_cur;
	struct tm tm_cur;
	arp_t *arp;
	
	/* Localtime */
	time_cur = time(NULL);
    	tm_cur = *(struct tm *) localtime(&time_cur);
	printf("  - DARPI Start at %d:%d:%d ...\n", tm_cur.tm_hour, tm_cur.tm_min, tm_cur.tm_sec);
	
	if ((arp = darpi_arp_cache_open()) == NULL)
		return (-1);
	
	/* Delete all Arp Cache entries */
	if (darpi_arp_cache_delete_entries(arp) < 0)
		goto exit_failure;
	
	/* DARPI Realtime */
	if (darpi_realtime(arp) < 0)
		goto exit_failure;
		

	darpi_arp_cache_close(arp);
	return (0);
	
	exit_failure:
		darpi_arp_cache_close(arp);
		return (-1);
}

/*
 * Open and create Arp cache tail
 */
static arp_t *
darpi_arp_cache_open()
{
	arp_t *arp;
	
	if ((arp = arp_open()) == NULL)
	{
		ERROR(strerror(errno));
		return (NULL);
	}
	
	if (arp_loop(arp, arp_cache_list_create, NULL) < 0)
	{
		arp_close(arp);
		return (NULL);
	}
			
	return (arp);
}

/*
 * Destroy Arp cache tail
 */
static void
darpi_arp_cache_close(arp_t *arp)
{

	arp_cache_list_destroy();
	arp_close(arp);
}

/*
 * Delete all found entries in the
 * Arp cache reading Arp cache tail
 */
static int
darpi_arp_cache_delete_entries(arp_t *arp)
{
	int i;
	struct arp_entry entry;

	printf("  - DARPI deletes these Arp Cache entries:\n");
		
	TAILQ_FOREACH(arp_cache_pos, &arp_cache_head, entries)
	{
			entry.arp_pa.addr_type = ADDR_TYPE_IP;
				
			for (i = 0; i < ETHER_ADDR_LEN; i++)
				entry.arp_ha.addr_eth.data[i] = arp_cache_pos->mac.octet[i];
			
			entry.arp_pa.addr_ip = arp_cache_pos->ip.s_addr;
	
			if (arp_delete(arp, &entry) < 0)
			{
				ERROR(strerror(errno));
				return (-1);
			}
			
			printf("  - Inet4: %15s  ->  Mac: %17s\n", 
			       inet_ntoa(arp_cache_pos->ip), ether_ntoa(&arp_cache_pos->mac));
	}
	
	return (0);
}

/*
 * Adds or refresh an entry in Arp cache
 * checked by DARPI during the Realtime execution
 */
static int
darpi_arp_cache_add_entry(struct in_addr *ip, struct ether_addr *mac)
{
	int entry_len;
	char *entry;

	entry_len = strlen(inet_ntoa(*ip)) + strlen(ether_ntoa(mac)) + 2;
			
	if ((entry = malloc(sizeof(char) * entry_len)) == NULL)
	{
		ERROR(strerror(errno));
		return (-1);
	}		
	sprintf(entry, "%s %s", inet_ntoa(*ip), ether_ntoa(mac));
			
	/* Add/Refresh Arp Cache entry */
	if (arp_cache_add(DARPI, darpi_arp, entry) < 0)
	{
		free(entry);
		return (-1);
	}
		
	free(entry);

	return (0);
}

/*
 * Delete an entry in the Arp cache
 * checked by DARPI during the Realtime execution
 */
static int
darpi_arp_cache_del_entry(char *arg)
{

	if (arp_cache_del(DARPI, darpi_arp, arg) < 0)
		return (-1);
		
	return (0);
}

/*
 * DARPI Realtime execution
 * Process all outbound/inbound packets
 */
static int
darpi_realtime(arp_t *arp)
{
	pcap_t *pcap;

	if ((pcap = darpi_realtime_open_packets()) == NULL)
		return (-1);
	
	printf("  - DARPI Realtime Protect actived:\n");
	
	/* Used by darpi_realtime_read_packets() */
	darpi_arp = arp;
	
	/* Realtime Protect */
	if (pcap_loop(pcap, -1, darpi_realtime_read_packets, NULL) < 0)
	{
		ERROR(pcap_geterr(pcap));
	
		darpi_cache_list_destroy();
		darpi_realtime_close_packets(pcap);
		return (-1);
	}
	
	/* Ctrl + C */
	signal(SIGINT, darpi_realtime_signal);
	signal(SIGTERM, darpi_realtime_signal);
	signal(SIGKILL, darpi_realtime_signal);
	
	darpi_cache_list_destroy();
	darpi_realtime_close_packets(pcap);
	return (0);
}

/*
 * If SIGINT signal (Ctrl + C) or SIGTERM/SIGKILL arrives,
 * it destroy Arp cache tail and then it exits
 */
static void
darpi_realtime_signal(int sig)
{

	darpi_cache_list_destroy();

	exit(EXIT_SUCCESS);
}

/*
 * Open pcap file descriptor
 */
static pcap_t *
darpi_realtime_open_packets()
{
	char *filter = "arp", errbuf[PCAP_ERRBUF_SIZE];
	struct bpf_program compiled_filter;
	pcap_t *pcap;
	
	if ((pcap = pcap_open_live(device, BUFSIZ, 0, 0, errbuf)) == NULL)
	{
		ERROR(errbuf);
		return (NULL);
	}

	if (pcap_setnonblock(pcap, 1, errbuf) < 0)
	{
		ERROR(pcap_geterr(pcap));
		goto exit_failure;
	}

	if (pcap_compile(pcap, &compiled_filter, filter, 0, device_netmask.s_addr) < 0)
	{		
		ERROR(pcap_geterr(pcap));
		goto exit_failure;
	}
	
	if (pcap_setfilter(pcap, &compiled_filter) < 0)
	{
		ERROR(pcap_geterr(pcap));
		goto exit_failure;
	}	
	
	pcap_freecode(&compiled_filter);

	return (pcap);
	
	exit_failure:
		darpi_realtime_close_packets(pcap);
		return (NULL);
}

/*
 * Close pcap file descriptor
 */
static void
darpi_realtime_close_packets(pcap_t *pcap)
{

	pcap_close(pcap);
}

/*
 * Reads inbound packets such as:
 *     - Arp request:
 *		Checks the source, if it matches us:
 *		    Save the destination address in the DARPI Cache
 *		    then it checks the inbound Arp reply packets
 *		Or checks the destination, if it matches us:
 *		    Delete source address in Arp cache
 *		    (possible address poisoned)
 *     - Arp reply:
 *		Checks the destination, if matches us:
 *		    Checks the source address, if it matches in
 *		    DARPI cache, they are accepted and inserted
 *		    in Arp cache, otherwise they are rejected
 *		    and deleted in Arp cache
 */
static void
darpi_realtime_read_packets(unsigned char *arg, const struct pcap_pkthdr *header, const unsigned char *packet)
{
	int i;
	char 	c_src_ip[INET4_ADDR_LEN], c_src_mac[MAC_ADDR_LEN],
		c_dst_ip[INET4_ADDR_LEN], c_dst_mac[MAC_ADDR_LEN];
	struct ether_addr src_mac, dst_mac;
	struct in_addr src_ip, dst_ip;
	struct arp_header *arp_pack;

	arp_pack = (struct arp_header *) (packet + device_offset);

	if (ntohs(arp_pack->header.ar_op) == ARP_OP_REQUEST)
	{
		sprintf(c_src_ip, "%d.%d.%d.%d", 
			arp_pack->addresses.ar_spa[0], arp_pack->addresses.ar_spa[1],
			arp_pack->addresses.ar_spa[2], arp_pack->addresses.ar_spa[3]);
		
		for (i = 0; i < ETHER_ADDR_LEN; i++)
			src_mac.octet[i] = arp_pack->addresses.ar_sha[i];
		strncpy(c_src_mac, ether_ntoa(&src_mac), sizeof(char) * MAC_ADDR_LEN);
		c_src_mac[sizeof(char) * strlen(c_src_mac)] = '\0';

		sprintf(c_dst_ip, "%d.%d.%d.%d", 
			arp_pack->addresses.ar_tpa[0], arp_pack->addresses.ar_tpa[1],
			arp_pack->addresses.ar_tpa[2], arp_pack->addresses.ar_tpa[3]);
		
		/* Check Source Arp request */
		if (strcmp(c_src_mac, ether_ntoa(&device_mac)) == 0 &&
		    strcmp(c_src_ip, inet_ntoa(device_inet4)) == 0)
		{
			inet_aton(c_dst_ip, &dst_ip);
			
			printf("  - DARPI ===== ARP REQUEST =====\n");
			
			/* DARPI Cache Add/Refresh */
			if (darpi_cache_list_create(&dst_ip) < 0)
				exit(EXIT_FAILURE);
		
			/* DARPI Cache Read */
			darpi_cache_list_print();
		}
		/* Check Destination Arp request */
		else if (strcmp(c_dst_ip, inet_ntoa(device_inet4)) == 0)
		{
			printf("  - DARPI ===== ARP REQUEST =====\n");

			/* Delete it in Arp cache (possible source address poisoned) */
			if (darpi_arp_cache_del_entry(c_src_ip) < 0)
				exit(EXIT_FAILURE);
		}
	}
	else if (ntohs(arp_pack->header.ar_op) == ARP_OP_REPLY)
	{
		sprintf(c_dst_ip, "%d.%d.%d.%d", 
			arp_pack->addresses.ar_tpa[0], arp_pack->addresses.ar_tpa[1],
			arp_pack->addresses.ar_tpa[2], arp_pack->addresses.ar_tpa[3]);
		
		for (i = 0; i < ETHER_ADDR_LEN; i++)
			dst_mac.octet[i] = arp_pack->addresses.ar_tha[i];
		strncpy(c_dst_mac, ether_ntoa(&dst_mac), sizeof(char) * MAC_ADDR_LEN);
		c_dst_mac[sizeof(char) * strlen(c_dst_mac)] = '\0';
	
		/* Check Destination Arp Reply */
		if (strcmp(c_dst_mac, ether_ntoa(&device_mac)) == 0 &&
		    strcmp(c_dst_ip, inet_ntoa(device_inet4)) == 0)
		{
			sprintf(c_src_ip, "%d.%d.%d.%d", 
				arp_pack->addresses.ar_spa[0], arp_pack->addresses.ar_spa[1],
				arp_pack->addresses.ar_spa[2], arp_pack->addresses.ar_spa[3]);
			inet_aton(c_src_ip, &src_ip);
			
			for (i = 0; i < ETHER_ADDR_LEN; i++)
				src_mac.octet[i] = arp_pack->addresses.ar_sha[i];
			strncpy(c_src_mac, ether_ntoa(&src_mac), sizeof(char) * MAC_ADDR_LEN);
			c_src_mac[sizeof(char) * strlen(c_src_mac)] = '\0';
	
			printf("  - DARPI ===== ARP REPLY =====\n");
			
			/* 
			 * Search entry in DARPI cache,
			 * if it matches it is to inserted in Arp cache
			 */
			if (darpi_cache_list_search(c_src_ip) == 0)
			{
				if (darpi_arp_cache_add_entry(&src_ip, &src_mac) < 0)
					exit(EXIT_FAILURE);
			}
			/*
			 * Entry don't exist in DARPI cache,
			 * it is to deleted in Arp cache
			 */
			else
			{
				if (darpi_arp_cache_del_entry(c_src_ip) < 0)
					exit(EXIT_FAILURE);
			}
		}
	}
}

/*
 * Create or adds DARPI cache tail
 */
static int
darpi_cache_list_create(struct in_addr *entry)
{
	struct in_addr ip;

	ip.s_addr = entry->s_addr;
	
	printf("  - DARPI Realtime adds %s entry in DARPI Cache!\n", inet_ntoa(*entry));
	
	if (darpi_cache_begin == NULL) 
	{
		TAILQ_INIT(&darpi_cache_head);
	
		if ((darpi_cache_begin = malloc(sizeof(struct darpi_cache))) == NULL)
		{
			ERROR(strerror(errno));
			return (-1);
		}
		memcpy(&darpi_cache_begin->ip, &ip, sizeof(ip));
	
		TAILQ_INSERT_HEAD(&darpi_cache_head, darpi_cache_begin, entries);
	}
	else
	{
		if ((darpi_cache_next = malloc(sizeof(struct darpi_cache))) == NULL)
		{
			ERROR(strerror(errno));
			return (-1);
		}
		memcpy(&darpi_cache_next->ip, &ip, sizeof(ip));

		TAILQ_INSERT_TAIL(&darpi_cache_head, darpi_cache_next, entries);
	}
	
	return (0);
}

/*
 * Prints DARPI cache
 */
static void
darpi_cache_list_print()
{
	unsigned int i = 1;

	printf("  - DARPI Cache list:\n");
	TAILQ_FOREACH(darpi_cache_pos, &darpi_cache_head, entries)
	{
		printf("  %d) Inet4: %15s\n", i, inet_ntoa(darpi_cache_pos->ip));
		
		i++;
	}
}

/*
 * Search entry in DARPI cache,
 * if it matches it will to deleted in DARPI cache
 */
static int
darpi_cache_list_search(char *entry)
{
	
	TAILQ_FOREACH(darpi_cache_pos, &darpi_cache_head, entries)
	{
		if (strcmp(inet_ntoa(darpi_cache_pos->ip), entry) == 0)
		{
			printf("  - DARPI Realtime %s entry found in DARPI Cache!\n", entry);
			printf("  - DARPI Realtime deletes %s entry in DARPI Cache!\n", entry);
			
			TAILQ_REMOVE(&darpi_cache_head, darpi_cache_pos, entries);
			free(darpi_cache_pos);
			
			return (0);
		}
	}
	
	printf("  - DARPI Realtime %s entry not found in DARPI Cache!\n", entry);
	
	return (-1);
}

/*
 * Destroy DARPI cache tail
 */
static void
darpi_cache_list_destroy()
{

	while (TAILQ_EMPTY(&darpi_cache_head) == 0)
	{
		darpi_cache_pos = TAILQ_FIRST(&darpi_cache_head);
		
		printf("  - DARPI Realtime deletes %s entry in DARPI Cache!\n", inet_ntoa(darpi_cache_pos->ip));
		
		TAILQ_REMOVE(&darpi_cache_head, darpi_cache_pos, entries);
		free(darpi_cache_pos);
     	}
}

/*
 * MISC handler:
 */

/*
 * Prints license
 */
static void
license()
{

	printf("  Copyright (c) 2008 Andrea Di Pasquale <spikey.it@gmail.com>\n");
 	printf("  All rights reserved.\n");
 	printf("\n");
 	printf("  Redistribution and use in source and binary forms, with or without\n");
 	printf("  modification, are permitted provided that the following conditions\n");
 	printf("  are met:\n");
 	printf("  1. Redistributions of source code must retain the above copyright\n");
 	printf("     notice(s), this list of conditions and the following disclaimer as\n");
 	printf("     the first lines of this file unmodified other than the possible\n");
 	printf("     addition of one or more copyright notices.\n");
 	printf("  2. Redistributions in binary form must reproduce the above copyright\n");
 	printf("     notice(s), this list of conditions and the following disclaimer in the\n");
 	printf("     documentation and/or other materials provided with the distribution.\n");
 	printf("\n");
 	printf("  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDER(S) ``AS IS'' AND ANY\n");
 	printf("  EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED\n");
 	printf("  WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE\n");
 	printf("  DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT HOLDER(S) BE LIABLE FOR ANY\n");
 	printf("  DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES\n");
 	printf("  (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR\n");
 	printf("  SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER\n");
 	printf("  CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT\n");
 	printf("  LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY\n");
 	printf("  OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH\n");
 	printf("  DAMAGE.\n");
}

/*
 * Prints version
 */
static void
version()
{

	printf("ArpON \"Arp handler inspectiON\" version %s (%s)\n", CUR_VERSION, WEB_SITE);
}

/*
 * Prints help summary page
 */
static void
help()
{

	version();
	
	printf("\n");
	printf(" Usage: arpon [Task Mode] {<Device + Arping> | <Device + Sniffer> | <Arp Cache> | <Deivce + <SARPI | DARPI>> | <Misc>}\n");
	printf("\n");	
	printf("TASK MODE:\n");
	printf("  --daemon,         -d                   Works in background task\n");
	printf("                                         (Default: foreground task)\n");
	printf("DEVICE MANAGER:\n");
	printf("  --dev-manual,     -i <device>          Sets your Ethernet device manually\n");
	printf("  --dev-auto,       -o                   Sets Ethernet device automatically\n");
	printf("  --dev-list,       -l                   Prints all Ethernet devices\n");
	printf("                                         (Sets last Ethernet device)\n");
	printf("ARP PING:\n");
	printf("  --ping-timeout,   -m <Timeout>         Sets Arp Ping response timeout\n");
	printf("                                         (Default: 500 ms)\n");
	printf("  --ping-host,      -p <Inet4>           Sends Arp Ping to Inet4 address\n");
	printf("  --ping-broadcast, -b                   Sends Arp Ping to Broadcast address\n");
	printf("                                         (Builds and prints LAN's active hosts)\n");
	printf("ARP PASSIVE SNIFFER:\n");
	printf("  --sniff-arp,      -r                   Sniffs only Arp protocol\n");
	printf("                                         (I/O Arp Request/Reply)\n");
	printf("ARP CACHE MANAGER:\n");
	printf("  --cache-add,      -a <\"Inet4  MAC\">    Adds Inet4 and MAC Arp entry\n");
	printf("  --cache-del,      -e <Inet4 | MAC>     Deletes Inet4 or MAC Arp entry\n");
	printf("  --cache-list,     -t                   Prints total ARP Cache entries\n");
	printf("\n");	
	printf("STATIC ARP INSPECTION:\n");
	printf("  --sarpi-timeout,  -u <Timeout>         Sets Arp Cache refresh timeout\n");
	printf("                                         (Default: 10 minuts)\n");
	printf("  --sarpi,          -s                   Manages Arp Cache statically\n");
	printf("\n");
	printf("DYNAMIC ARP INSPECTION:\n");
	printf("  --darpi,          -y                   Manages Arp Cache dinamically\n");
	printf("\n");	
	printf("MISC:\n");
	printf("  --license,        -c                   Prints license page\n");
	printf("  --version,        -v                   Prints version number\n");
	printf("  --help,           -h                   Prints help summary page\n");
	printf("\n");	
	printf("SEE THE MAN PAGE FOR MANY DESCRIPTIONS AND EXAMPLES\n");
}

/*
 * MAIN handler:
 */

int
main(int argc, char *argv[])
{
	register int gopt;
	struct option longopts[] = {
		{"daemon",		no_argument,		NULL,	'd' },
		
		{"dev-manual",		required_argument,	NULL,   'i' },
		{"dev-auto",		no_argument,		NULL,	'o' },
		{"dev-list",		no_argument,        	NULL,	'l' },

		{"ping-timeout",	required_argument,	NULL,	'm' },	
		{"ping-host",		required_argument,	NULL,   'p' },
		{"ping-broadcast",	no_argument,		NULL,	'b' },
	
		{"sniff-arp",		no_argument,		NULL,	'r' },
	
		{"cache-add",		required_argument,  	NULL,	'a' },
		{"cache-del",		required_argument,  	NULL,	'e' },
		{"cache-list",		no_argument,		NULL,	't' },
	
		{"sarpi-timeout",	required_argument,	NULL,	'u' },
		{"sarpi",		no_argument,		NULL,	's' },
		
		{"darpi",		no_argument,		NULL,	'y' },
	
		{"license",		no_argument,		NULL,	'c' },
		{"version",		no_argument,		NULL,	'v' },
		{"help",		no_argument,		NULL,	'h' },
	
		{NULL,			0,                  	NULL,	 0  }
	};

	printf("\n");
	
	if (argc == 1) 
	{
		help();
		goto exit_success;
	}
	
	if (getuid() != ROOT_USER)
	{
		fprintf(stderr, "ERROR: You are not root user!\n");
		goto exit_failure;
	}
	
	while ((gopt = getopt_long(argc, argv, "di:olm:p:bra:e:tu:sycvh", longopts, NULL)) != -1) 
	{
		switch(gopt)
		{
			case ('d'):
				if (task_mode_daemon() < 0)
					goto exit_failure;
				break;
				
			case ('i'):
				if (device_manager(DEVICE_MANUAL, optarg) < 0)
					goto exit_failure;
				break;
				
			case ('o'):
				if (device_manager(DEVICE_AUTO, NULL) < 0)
					goto exit_failure;
				break;
				
			case ('l'):
				if (device_manager(DEVICE_LIST, NULL) < 0)
					goto exit_failure;
				break;
		
			case ('m'):
				if (arp_ping_set_timeout(optarg) < 0)
					goto exit_failure;
				break;

			case ('p'):
				if (device[0] == '\0')
				{
					fprintf(stderr, "ERROR: you must to set device!\n");
					goto exit_failure;
				}
				
				if (arp_ping_host(optarg) < 0)
					goto exit_failure;
				break;
				
			case ('b'):
				if (device[0] == '\0')
				{
					fprintf(stderr, "ERROR: you must to set device!\n");
					goto exit_failure;
				}
				
				if (arp_ping_broadcast() < 0)
					goto exit_failure;
				break;
				
			case ('r'):
				if (device[0] == '\0')
				{
					fprintf(stderr, "ERROR: you must to set device!\n");
					goto exit_failure;
				}
				
				if (arp_sniff_packets() < 0)
					goto exit_failure;
				break;
				
			case ('a'):
				if (arp_cache_manager(ARP_CACHE_ADD, optarg) < 0)
					goto exit_failure;
				break;
			
			case ('e'):
				if (arp_cache_manager(ARP_CACHE_DEL, optarg) < 0)
					goto exit_failure;
				break;
			
			case ('t'):
				if (arp_cache_manager(ARP_CACHE_LIST, NULL) < 0)
					goto exit_failure;
				break;
				
			case ('u'):
				sarpi_set_timeout(optarg);
				break;
				
			case ('s'):
				if (device[0] == '\0')
				{
					fprintf(stderr, "ERROR: you must to set device!\n");
					goto exit_failure;
				}
				
				if (sarpi_manager() < 0)
					goto exit_failure;
				break;
				
			case ('y'):
				if (device[0] == '\0')
				{
					fprintf(stderr, "ERROR: you must to set device!\n");
					goto exit_failure;
				}
				
				if (darpi_manager() < 0)
					goto exit_failure;
				break;
				
			case ('c'):
				license();
				goto exit_success;
	
			case ('v'):
				version();
				goto exit_success;
	
			case ('h'):
				help();
				goto exit_success;
	
			default:
				break;
		}
	}
	
	argc -= optind;
	argv += optind;
	 
	printf("\n");
	return (0);
	
	exit_failure:
		printf("\n");
		return (-1);

	exit_success:
		printf("\n");
		return (0);
}

/* EOF */
